using System;
using System.ComponentModel;
using System.Collections;
using System.Diagnostics;
using System.Drawing;
using System.Windows.Forms;
using System.Runtime.InteropServices;
using System.Drawing.Imaging;
using System.IO;
using System.Collections.Specialized;
using System.Text;


namespace HtmlRichText
{
    #region Public Enums

    // Enum for possible RTF colors
    public enum RtfColor
    {
        Black, Maroon, Green, Olive, Navy, Purple, Teal, Gray, Silver,
        Red, Lime, Yellow, Blue, Fuchsia, Aqua, White
    }

    #endregion

    public class HtmlRichTextBox : RichTextBox
    {
        private int updating = 0;
        private int oldEventMask = 0;

        /// <summary>
        /// Maintains performance while updating.
        /// </summary>
        /// <remarks>
        /// <para>
        /// It is recommended to call this method before doing
        /// any major updates that you do not wish the user to
        /// see. Remember to call EndUpdate when you are finished
        /// with the update. Nested calls are supported.
        /// </para>
        /// <para>
        /// Calling this method will prevent redrawing. It will
        /// also setup the event mask of the underlying richedit
        /// control so that no events are sent.
        /// </para>
        /// </remarks>

        public void BeginUpdate()
        {
            // Deal with nested calls.
            ++updating;

            if (updating > 1)
                return;

            // Prevent the control from raising any events.
            oldEventMask = SendMessage(new HandleRef(this, Handle),
                EM_SETEVENTMASK, 0, 0);

            // Prevent the control from redrawing itself.
            SendMessage(new HandleRef(this, Handle),
                WM_SETREDRAW, 0, 0);
        }

        /// <summary>
        /// Resumes drawing and event handling.
        /// </summary>
        /// <remarks>
        /// This method should be called every time a call is made
        /// made to BeginUpdate. It resets the event mask to it's
        /// original value and enables redrawing of the control.
        /// </remarks>
        public void EndUpdate()
        {
            // Deal with nested calls.
            --updating;

            if (updating > 0)
                return;

            // Allow the control to redraw itself.
            SendMessage(new HandleRef(this, Handle),
                WM_SETREDRAW, 1, 0);

            // Allow the control to raise event messages.
            SendMessage(new HandleRef(this, Handle),
                EM_SETEVENTMASK, 0, oldEventMask);
        }

        /// <summary>
        /// Returns true when the control is performing some 
        /// internal updates, specially when is reading or writing
        /// HTML text
        /// </summary>
        public bool InternalUpdating
        {
            get
            {
                return (updating != 0);
            }
        }

        #region Win32 Apis

        // Constants from the Platform SDK.
        private const int EM_FORMATRANGE = 1081;

        private const int WM_USER = 0x0400;
        private const int EM_GETCHARFORMAT = WM_USER + 58;
        private const int EM_SETCHARFORMAT = WM_USER + 68;

        private const int EM_SETEVENTMASK = 1073;
        private const int EM_GETPARAFORMAT = 1085;
        private const int EM_SETPARAFORMAT = 1095;
        private const int EM_SETTYPOGRAPHYOPTIONS = 1226;
        private const int WM_SETREDRAW = 11;
        private const int TO_ADVANCEDTYPOGRAPHY = 1;


        // Defines for EM_SETCHARFORMAT/EM_GETCHARFORMAT
        private const Int32 SCF_SELECTION = 0x0001;
        private const Int32 SCF_WORD = 0x0002;
        private const Int32 SCF_ALL = 0x0004;

        public const int LF_FACESIZE = 32;

        // Defines for STRUCT_CHARFORMAT member dwMask
        public const UInt32 CFM_BOLD = 0x00000001;
        public const UInt32 CFM_ITALIC = 0x00000002;
        public const UInt32 CFM_UNDERLINE = 0x00000004;
        public const UInt32 CFM_STRIKEOUT = 0x00000008;
        public const UInt32 CFM_PROTECTED = 0x00000010;
        public const UInt32 CFM_LINK = 0x00000020;
        public const UInt32 CFM_SIZE = 0x80000000;
        public const UInt32 CFM_COLOR = 0x40000000;
        public const UInt32 CFM_FACE = 0x20000000;
        public const UInt32 CFM_OFFSET = 0x10000000;
        public const UInt32 CFM_CHARSET = 0x08000000;
        public const UInt32 CFM_SUBSCRIPT = CFE_SUBSCRIPT | CFE_SUPERSCRIPT;
        public const UInt32 CFM_SUPERSCRIPT = CFM_SUBSCRIPT;

        // Defines for STRUCT_CHARFORMAT member dwEffects
        public const UInt32 CFE_BOLD = 0x00000001;
        public const UInt32 CFE_ITALIC = 0x00000002;
        public const UInt32 CFE_UNDERLINE = 0x00000004;
        public const UInt32 CFE_STRIKEOUT = 0x00000008;
        public const UInt32 CFE_PROTECTED = 0x00000010;
        public const UInt32 CFE_LINK = 0x00000020;
        public const UInt32 CFE_AUTOCOLOR = 0x40000000;
        public const UInt32 CFE_SUBSCRIPT = 0x00010000;		/* Superscript and subscript are */
        public const UInt32 CFE_SUPERSCRIPT = 0x00020000;		/*  mutually exclusive			 */

        public const byte CFU_UNDERLINENONE = 0x00;
        public const byte CFU_UNDERLINE = 0x01;
        public const byte CFU_UNDERLINEWORD = 0x02; /* (*) displayed as ordinary underline	*/
        public const byte CFU_UNDERLINEDOUBLE = 0x03; /* (*) displayed as ordinary underline	*/
        public const byte CFU_UNDERLINEDOTTED = 0x04;
        public const byte CFU_UNDERLINEDASH = 0x05;
        public const byte CFU_UNDERLINEDASHDOT = 0x06;
        public const byte CFU_UNDERLINEDASHDOTDOT = 0x07;
        public const byte CFU_UNDERLINEWAVE = 0x08;
        public const byte CFU_UNDERLINETHICK = 0x09;
        public const byte CFU_UNDERLINEHAIRLINE = 0x0A; /* (*) displayed as ordinary underline	*/

        public const int CFM_SMALLCAPS = 0x0040;			/* (*)	*/
        public const int CFM_ALLCAPS = 0x0080;			/* Displayed by 3.0	*/
        public const int CFM_HIDDEN = 0x0100;			/* Hidden by 3.0 */
        public const int CFM_OUTLINE = 0x0200;			/* (*)	*/
        public const int CFM_SHADOW = 0x0400;			/* (*)	*/
        public const int CFM_EMBOSS = 0x0800;			/* (*)	*/
        public const int CFM_IMPRINT = 0x1000;			/* (*)	*/
        public const int CFM_DISABLED = 0x2000;
        public const int CFM_REVISED = 0x4000;

        public const int CFM_BACKCOLOR = 0x04000000;
        public const int CFM_LCID = 0x02000000;
        public const int CFM_UNDERLINETYPE = 0x00800000;		/* Many displayed by 3.0 */
        public const int CFM_WEIGHT = 0x00400000;
        public const int CFM_SPACING = 0x00200000;		/* Displayed by 3.0	*/
        public const int CFM_KERNING = 0x00100000;		/* (*)	*/
        public const int CFM_STYLE = 0x00080000;		/* (*)	*/
        public const int CFM_ANIMATION = 0x00040000;		/* (*)	*/
        public const int CFM_REVAUTHOR = 0x00008000;

        // Font Weights
        public const short FW_DONTCARE = 0;
        public const short FW_THIN = 100;
        public const short FW_EXTRALIGHT = 200;
        public const short FW_LIGHT = 300;
        public const short FW_NORMAL = 400;
        public const short FW_MEDIUM = 500;
        public const short FW_SEMIBOLD = 600;
        public const short FW_BOLD = 700;
        public const short FW_EXTRABOLD = 800;
        public const short FW_HEAVY = 900;

        public const short FW_ULTRALIGHT = FW_EXTRALIGHT;
        public const short FW_REGULAR = FW_NORMAL;
        public const short FW_DEMIBOLD = FW_SEMIBOLD;
        public const short FW_ULTRABOLD = FW_EXTRABOLD;
        public const short FW_BLACK = FW_HEAVY;

        // PARAFORMAT mask values
        public const UInt32 PFM_STARTINDENT = 0x00000001;
        public const UInt32 PFM_RIGHTINDENT = 0x00000002;
        public const UInt32 PFM_OFFSET = 0x00000004;
        public const UInt32 PFM_ALIGNMENT = 0x00000008;
        public const UInt32 PFM_TABSTOPS = 0x00000010;
        public const UInt32 PFM_NUMBERING = 0x00000020;
        public const UInt32 PFM_OFFSETINDENT = 0x80000000;

        // PARAFORMAT numbering options
        public const UInt16 PFN_BULLET = 0x0001;

        // PARAFORMAT alignment options
        public const UInt16 PFA_LEFT = 0x0001;
        public const UInt16 PFA_RIGHT = 0x0002;
        public const UInt16 PFA_CENTER = 0x0003;



        // It makes no difference if we use PARAFORMAT or
        // PARAFORMAT2 here, so I have opted for PARAFORMAT2.
        [StructLayout(LayoutKind.Sequential)]
        public struct PARAFORMAT
        {
            public int cbSize;
            public uint dwMask;
            public short wNumbering;
            public short wReserved;
            public int dxStartIndent;
            public int dxRightIndent;
            public int dxOffset;
            public short wAlignment;
            public short cTabCount;
            [MarshalAs(UnmanagedType.ByValArray, SizeConst = 32)]
            public int[] rgxTabs;

            // PARAFORMAT2 from here onwards.
            public int dySpaceBefore;
            public int dySpaceAfter;
            public int dyLineSpacing;
            public short sStyle;
            public byte bLineSpacingRule;
            public byte bOutlineLevel;
            public short wShadingWeight;
            public short wShadingStyle;
            public short wNumberingStart;
            public short wNumberingStyle;
            public short wNumberingTab;
            public short wBorderSpace;
            public short wBorderWidth;
            public short wBorders;
        }

        [StructLayout(LayoutKind.Sequential)]
        public struct CHARFORMAT
        {
            public int cbSize;
            public UInt32 dwMask;
            public UInt32 dwEffects;
            public Int32 yHeight;
            public Int32 yOffset;
            public Int32 crTextColor;
            public byte bCharSet;
            public byte bPitchAndFamily;
            [MarshalAs(UnmanagedType.ByValArray, SizeConst = 32)]
            public char[] szFaceName;

            // CHARFORMAT2 from here onwards.
            public short wWeight;
            public short sSpacing;
            public Int32 crBackColor;
            public uint lcid;
            public uint dwReserved;
            public short sStyle;
            public short wKerning;
            public byte bUnderlineType;
            public byte bAnimation;
            public byte bRevAuthor;
            public byte bReserved1;
        }

        [DllImport("user32", CharSet = CharSet.Auto)]
        private static extern int SendMessage(HandleRef hWnd,
            int msg,
            int wParam,
            int lParam);

        [DllImport("user32", CharSet = CharSet.Auto)]
        private static extern int SendMessage(HandleRef hWnd,
            int msg,
            int wParam,
            ref PARAFORMAT lp);

        [DllImport("user32", CharSet = CharSet.Auto)]
        private static extern int SendMessage(HandleRef hWnd,
            int msg,
            int wParam,
            ref CHARFORMAT lp);

        #endregion

        //----------------------------
        // New, version 1.1
        public void SetSuperScript(bool bSet)
        {
            CHARFORMAT cf = this.CharFormat;

            if (bSet)
            {
                cf.dwMask |= CFM_SUPERSCRIPT;
                cf.dwEffects |= CFE_SUPERSCRIPT;
            }
            else
            {
                cf.dwEffects &= ~CFE_SUPERSCRIPT;
            }

            this.CharFormat = cf;
        }

        public void SetSubScript(bool bSet)
        {
            CHARFORMAT cf = this.CharFormat;

            if (bSet)
            {
                cf.dwMask |= CFM_SUBSCRIPT;
                cf.dwEffects |= CFE_SUBSCRIPT;
            }
            else
            {
                cf.dwEffects &= ~CFE_SUBSCRIPT;
            }

            this.CharFormat = cf;
        }

        public bool IsSuperScript()
        {
            CHARFORMAT cf = this.CharFormat;
            return ((cf.dwEffects & CFE_SUPERSCRIPT) == CFE_SUPERSCRIPT);
        }

        public bool IsSubScript()
        {
            CHARFORMAT cf = this.CharFormat;
            return ((cf.dwEffects & CFE_SUBSCRIPT) == CFE_SUBSCRIPT);
        }
        //----------------------------

        public PARAFORMAT ParaFormat
        {
            get
            {
                PARAFORMAT pf = new PARAFORMAT();
                pf.cbSize = Marshal.SizeOf(pf);

                // Get the alignment.
                SendMessage(new HandleRef(this, Handle),
                    EM_GETPARAFORMAT,
                    SCF_SELECTION, ref pf);

                return pf;
            }

            set
            {
                PARAFORMAT pf = value;
                pf.cbSize = Marshal.SizeOf(pf);

                // Set the alignment.
                SendMessage(new HandleRef(this, Handle),
                    EM_SETPARAFORMAT,
                    SCF_SELECTION, ref pf);
            }
        }

        public PARAFORMAT DefaultParaFormat
        {
            get
            {
                PARAFORMAT pf = new PARAFORMAT();
                pf.cbSize = Marshal.SizeOf(pf);

                // Get the alignment.
                SendMessage(new HandleRef(this, Handle),
                    EM_GETPARAFORMAT,
                    SCF_ALL, ref pf);

                return pf;
            }

            set
            {
                PARAFORMAT pf = value;
                pf.cbSize = Marshal.SizeOf(pf);

                // Set the alignment.
                SendMessage(new HandleRef(this, Handle),
                    EM_SETPARAFORMAT,
                    SCF_ALL, ref pf);
            }
        }

        public CHARFORMAT CharFormat
        {
            get
            {
                CHARFORMAT cf = new CHARFORMAT();
                cf.cbSize = Marshal.SizeOf(cf);

                // Get the alignment.
                SendMessage(new HandleRef(this, Handle),
                    EM_GETCHARFORMAT,
                    SCF_SELECTION, ref cf);

                return cf;
            }

            set
            {
                CHARFORMAT cf = value;
                cf.cbSize = Marshal.SizeOf(cf);

                // Set the alignment.
                SendMessage(new HandleRef(this, Handle),
                    EM_SETCHARFORMAT,
                    SCF_SELECTION, ref cf);
            }
        }

        public CHARFORMAT DefaultCharFormat
        {
            get
            {
                CHARFORMAT cf = new CHARFORMAT();
                cf.cbSize = Marshal.SizeOf(cf);

                // Get the alignment.
                SendMessage(new HandleRef(this, Handle),
                    EM_GETCHARFORMAT,
                    SCF_ALL, ref cf);

                return cf;
            }

            set
            {
                CHARFORMAT cf = value;
                cf.cbSize = Marshal.SizeOf(cf);

                // Set the alignment.
                SendMessage(new HandleRef(this, Handle),
                    EM_SETCHARFORMAT,
                    SCF_ALL, ref cf);
            }
        }

        #region COLORREF helper functions

        // convert COLORREF to Color
        private Color GetColor(int crColor)
        {
            byte r = (byte)(crColor);
            byte g = (byte)(crColor >> 8);
            byte b = (byte)(crColor >> 16);

            return Color.FromArgb(r, g, b);
        }

        // convert COLORREF to Color
        private int GetCOLORREF(int r, int g, int b)
        {
            int r2 = r;
            int g2 = (g << 8);
            int b2 = (b << 16);

            int result = r2 | g2 | b2;

            return result;
        }

        private int GetCOLORREF(Color color)
        {
            int r = color.R;
            int g = color.G;
            int b = color.B;

            return GetCOLORREF(r, g, b);
        }
        #endregion


        #region "Get HTML text"

        // format states
        private enum ctformatStates
        {
            nctNone = 0, // none format applied
            nctNew = 1, // new format
            nctContinue = 2, // continue with previous format
            nctReset = 3 // reset format (close this tag)
        }

        private enum uMyREType
        {
            U_MYRE_TYPE_TAG,
            U_MYRE_TYPE_EMO,
            U_MYRE_TYPE_ENTITY,
        }

        private struct cMyREFormat
        {
            public uMyREType nType;
            public int nLen;
            public int nPos;
            public string strValue;
        }

        public string GetHTML(bool bHTML, bool bParaFormat)
        {
            //------------------------
            // working variables
            CHARFORMAT cf;
            PARAFORMAT pf;

            ctformatStates bold = ctformatStates.nctNone;
            ctformatStates bitalic = ctformatStates.nctNone;
            ctformatStates bstrikeout = ctformatStates.nctNone;
            ctformatStates bunderline = ctformatStates.nctNone;
            ctformatStates super = ctformatStates.nctNone;
            ctformatStates sub = ctformatStates.nctNone;

            ctformatStates bacenter = ctformatStates.nctNone;
            ctformatStates baleft = ctformatStates.nctNone;
            ctformatStates baright = ctformatStates.nctNone;
            ctformatStates bnumbering = ctformatStates.nctNone;

            string strFont = "";
            Int32 crFont = 0;
            Color color = new Color();
            int yHeight = 0;
            int i = 0;
            //-------------------------

            //-------------------------
            // to store formatting
            ArrayList colFormat = new ArrayList();

            cMyREFormat mfr;
            //-------------------------

            //-------------------------
            // ok, lets go
            int nStart, nEnd;
            string strHTML = "";

            this.HideSelection = true;
            this.BeginUpdate();

            nStart = this.SelectionStart;
            nEnd = this.SelectionLength;

            try
            {
                //--------------------------------
                // replace entities
                if (bHTML)
                {

                    char[] ch = { '&', '<', '>', '"', '\'' };
                    string[] strreplace = { "&amp;", "&lt;", "&gt;", "&quot;", "&apos;" };

                    for (i = 0; i < ch.Length; i++)
                    {
                        char[] ch2 = { ch[i] };

                        int n = this.Find(ch2, 0);
                        while (n != -1)
                        {
                            mfr = new cMyREFormat();

                            mfr.nPos = n;
                            mfr.nLen = 1;
                            mfr.nType = uMyREType.U_MYRE_TYPE_ENTITY;
                            mfr.strValue = strreplace[i];

                            colFormat.Add(mfr);

                            n = this.Find(ch2, n + 1);
                        }
                    }
                }
                //--------------------------------

                string strT = "";

                int k = this.TextLength;
                char[] chtrim = { ' ', '\x0000' };

                //--------------------------------
                // this is an inefficient method to get text format
                // but RichTextBox doesn't provide another method to
                // get something like an array of charformat and paraformat
                //--------------------------------
                for (i = 0; i < k; i++)
                {
                    // select one character
                    this.Select(i, 1);
                    string strChar = this.SelectedText;

                    if (bHTML)
                    {
                        //-------------------------
                        // get format for this character
                        cf = this.CharFormat;
                        pf = this.ParaFormat;


                        if (this.SelectedRtf.Contains(@"{\pict\wmetafile8"))
                        {
                            BitmapBase64 Base64 = new BitmapBase64();
                            string sel = Base64.FixBase64ForImage(this.SelectedRtf);
                            for (int t = 0; t < images.Count; t++)
                            {
                                Base64 = (BitmapBase64)images[t];
                                if (sel.Contains(Base64.RTFstring))
                                {
                                    mfr = new cMyREFormat();

                                    mfr.nPos = i;
                                    mfr.nLen = 0;
                                    mfr.nType = uMyREType.U_MYRE_TYPE_TAG;
                                    mfr.strValue = "<img alt=\"missing image\" src=\"data:image/" + Base64.extension + ";base64," + Base64.Base64string + "\" style=\"border: 0px none ;\" />";

                                    colFormat.Add(mfr);
                                }
                            }
                        }
                        string strfname = new string(cf.szFaceName);
                        strfname = strfname.Trim(chtrim);
                        //-------------------------


                        //-------------------------
                        // new font format ?
                        if ((strFont != strfname) || (crFont != cf.crTextColor) || (yHeight != cf.yHeight))
                        {
                            if (strFont != "")
                            {
                                // close previous <font> tag

                                mfr = new cMyREFormat();

                                mfr.nPos = i;
                                mfr.nLen = 0;
                                mfr.nType = uMyREType.U_MYRE_TYPE_TAG;
                                mfr.strValue = "</font>";

                                colFormat.Add(mfr);
                            }

                            //-------------------------
                            // save this for cache
                            strFont = strfname;
                            crFont = cf.crTextColor;
                            yHeight = cf.yHeight;
                            //-------------------------

                            //-------------------------
                            // font size should be translate to 
                            // html size (Approximately)
                            int fsize = yHeight / (20 * 5);
                            //-------------------------

                            //-------------------------
                            // color object from COLORREF
                            color = GetColor(crFont);
                            //-------------------------

                            //-------------------------
                            // add <font> tag
                            mfr = new cMyREFormat();

                            string strcolor = string.Concat("#", (color.ToArgb() & 0x00FFFFFF).ToString("X6"));

                            mfr.nPos = i;
                            mfr.nLen = 0;
                            mfr.nType = uMyREType.U_MYRE_TYPE_TAG;
                            string mfrFont = "";
                            ///mfr.strValue = "<font face=\"" + strFont + "\" color=\"" + strcolor + "\" size=\"" + fsize + "\">";

                            if (strFont != this.Font.FontFamily.Name)
                            {
                                mfrFont += " face=\"" + strFont + "\"";
                            }
                            if (strcolor != "#000000")
                            {
                                mfrFont += " color=\"" + strcolor + "\"";
                            }
                            if (fsize != 1)
                            {
                                mfrFont += " size=\"" + fsize + "\"";
                            }
                            if (mfrFont != "")
                            {
                                mfr.strValue = "<font " + mfrFont + ">";
                            }
                            else
                            {
                                mfr.strValue = "";
                                strFont = "";
                            }
                            colFormat.Add(mfr);
                            //-------------------------
                        }

                        //-------------------------
                        // are we in another line ?
                        if ((strChar == "\r") || (strChar == "\n"))
                        {
                            // yes?
                            // then, we need to reset paragraph format
                            // and character format
                            if (bParaFormat)
                            {
                                bnumbering = ctformatStates.nctNone;
                                baleft = ctformatStates.nctNone;
                                baright = ctformatStates.nctNone;
                                bacenter = ctformatStates.nctNone;
                            }

                            // close previous tags

                            // is italic? => close it
                            if (bitalic != ctformatStates.nctNone)
                            {
                                mfr = new cMyREFormat();

                                mfr.nPos = i;
                                mfr.nLen = 0;
                                mfr.nType = uMyREType.U_MYRE_TYPE_TAG;
                                mfr.strValue = "</i>";

                                colFormat.Add(mfr);

                                bitalic = ctformatStates.nctNone;
                            }

                            // is bold? => close it
                            if (bold != ctformatStates.nctNone)
                            {
                                mfr = new cMyREFormat();

                                mfr.nPos = i;
                                mfr.nLen = 0;
                                mfr.nType = uMyREType.U_MYRE_TYPE_TAG;
                                mfr.strValue = "</b>";

                                colFormat.Add(mfr);

                                bold = ctformatStates.nctNone;
                            }

                            // is underline? => close it
                            if (bunderline != ctformatStates.nctNone)
                            {
                                mfr = new cMyREFormat();

                                mfr.nPos = i;
                                mfr.nLen = 0;
                                mfr.nType = uMyREType.U_MYRE_TYPE_TAG;
                                mfr.strValue = "</u>";

                                colFormat.Add(mfr);

                                bunderline = ctformatStates.nctNone;
                            }

                            // is strikeout? => close it
                            if (bstrikeout != ctformatStates.nctNone)
                            {
                                mfr = new cMyREFormat();

                                mfr.nPos = i;
                                mfr.nLen = 0;
                                mfr.nType = uMyREType.U_MYRE_TYPE_TAG;
                                mfr.strValue = "</s>";

                                colFormat.Add(mfr);

                                bstrikeout = ctformatStates.nctNone;
                            }

                            // is super? => close it
                            if (super != ctformatStates.nctNone)
                            {
                                mfr = new cMyREFormat();

                                mfr.nPos = i;
                                mfr.nLen = 0;
                                mfr.nType = uMyREType.U_MYRE_TYPE_TAG;
                                mfr.strValue = "</sup>";

                                colFormat.Add(mfr);

                                super = ctformatStates.nctNone;
                            }

                            // is sub? => close it
                            if (sub != ctformatStates.nctNone)
                            {
                                mfr = new cMyREFormat();

                                mfr.nPos = i;
                                mfr.nLen = 0;
                                mfr.nType = uMyREType.U_MYRE_TYPE_TAG;
                                mfr.strValue = "</sub>";

                                colFormat.Add(mfr);

                                sub = ctformatStates.nctNone;
                            }
                        }

                        // now, process the paragraph format,
                        // managing states: none, new, continue {with previous}, reset
                        if (bParaFormat)
                        {
                            // align to center?
                            if (pf.wAlignment == PFA_CENTER)
                            {
                                if (bacenter == ctformatStates.nctNone)
                                    bacenter = ctformatStates.nctNew;
                                else
                                    bacenter = ctformatStates.nctContinue;
                            }
                            else
                            {
                                if (bacenter != ctformatStates.nctNone)
                                    bacenter = ctformatStates.nctReset;
                            }

                            if (bacenter == ctformatStates.nctNew)
                            {
                                mfr = new cMyREFormat();

                                mfr.nPos = i;
                                mfr.nLen = 0;
                                mfr.nType = uMyREType.U_MYRE_TYPE_TAG;
                                mfr.strValue = "<p align=\"center\">";

                                colFormat.Add(mfr);
                            }
                            else if (bacenter == ctformatStates.nctReset)
                                bacenter = ctformatStates.nctNone;
                            //---------------------

                            //---------------------
                            // align to left ?
                            if (pf.wAlignment == PFA_LEFT)
                            {
                                if (baleft == ctformatStates.nctNone)
                                    baleft = ctformatStates.nctNew;
                                else
                                    baleft = ctformatStates.nctContinue;
                            }
                            else
                            {
                                if (baleft != ctformatStates.nctNone)
                                    baleft = ctformatStates.nctReset;
                            }

                            if (baleft == ctformatStates.nctNew)
                            {
                                mfr = new cMyREFormat();

                                mfr.nPos = i;
                                mfr.nLen = 0;
                                mfr.nType = uMyREType.U_MYRE_TYPE_TAG;
                                mfr.strValue = "<p align=\"left\">";

                                colFormat.Add(mfr);
                            }
                            else if (baleft == ctformatStates.nctReset)
                                baleft = ctformatStates.nctNone;
                            //---------------------

                            //---------------------
                            // align to right ?
                            if (pf.wAlignment == PFA_RIGHT)
                            {
                                if (baright == ctformatStates.nctNone)
                                    baright = ctformatStates.nctNew;
                                else
                                    baright = ctformatStates.nctContinue;
                            }
                            else
                            {
                                if (baright != ctformatStates.nctNone)
                                    baright = ctformatStates.nctReset;
                            }

                            if (baright == ctformatStates.nctNew)
                            {
                                mfr = new cMyREFormat();

                                mfr.nPos = i;
                                mfr.nLen = 0;
                                mfr.nType = uMyREType.U_MYRE_TYPE_TAG;
                                mfr.strValue = "<p align=\"right\">";

                                colFormat.Add(mfr);
                            }
                            else if (baright == ctformatStates.nctReset)
                                baright = ctformatStates.nctNone;
                            //---------------------

                            //---------------------
                            // bullet ?
                            if (pf.wNumbering == PFN_BULLET)
                            {
                                if (bnumbering == ctformatStates.nctNone)
                                    bnumbering = ctformatStates.nctNew;
                                else
                                    bnumbering = ctformatStates.nctContinue;
                            }
                            else
                            {
                                if (bnumbering != ctformatStates.nctNone)
                                    bnumbering = ctformatStates.nctReset;
                            }

                            if (bnumbering == ctformatStates.nctNew)
                            {
                                mfr = new cMyREFormat();

                                mfr.nPos = i;
                                mfr.nLen = 0;
                                mfr.nType = uMyREType.U_MYRE_TYPE_TAG;
                                mfr.strValue = "<li>";

                                colFormat.Add(mfr);
                            }
                            else if (bnumbering == ctformatStates.nctReset)
                                bnumbering = ctformatStates.nctNone;
                            //---------------------
                        }

                        //---------------------
                        // bold ?
                        if ((cf.dwEffects & CFE_BOLD) == CFE_BOLD)
                        {
                            if (bold == ctformatStates.nctNone)
                                bold = ctformatStates.nctNew;
                            else
                                bold = ctformatStates.nctContinue;
                        }
                        else
                        {
                            if (bold != ctformatStates.nctNone)
                                bold = ctformatStates.nctReset;
                        }

                        if (bold == ctformatStates.nctNew)
                        {
                            mfr = new cMyREFormat();

                            mfr.nPos = i;
                            mfr.nLen = 0;
                            mfr.nType = uMyREType.U_MYRE_TYPE_TAG;
                            mfr.strValue = "<b>";

                            colFormat.Add(mfr);
                        }
                        else if (bold == ctformatStates.nctReset)
                        {
                            mfr = new cMyREFormat();

                            mfr.nPos = i;
                            mfr.nLen = 0;
                            mfr.nType = uMyREType.U_MYRE_TYPE_TAG;
                            mfr.strValue = "</b>";

                            colFormat.Add(mfr);

                            bold = ctformatStates.nctNone;
                        }
                        //---------------------


                        //---------------------
                        // Italic
                        if ((cf.dwEffects & CFE_ITALIC) == CFE_ITALIC)
                        {
                            if (bitalic == ctformatStates.nctNone)
                                bitalic = ctformatStates.nctNew;
                            else
                                bitalic = ctformatStates.nctContinue;
                        }
                        else
                        {
                            if (bitalic != ctformatStates.nctNone)
                                bitalic = ctformatStates.nctReset;
                        }

                        if (bitalic == ctformatStates.nctNew)
                        {
                            mfr = new cMyREFormat();

                            mfr.nPos = i;
                            mfr.nLen = 0;
                            mfr.nType = uMyREType.U_MYRE_TYPE_TAG;
                            mfr.strValue = "<i>";

                            colFormat.Add(mfr);
                        }
                        else if (bitalic == ctformatStates.nctReset)
                        {
                            mfr = new cMyREFormat();

                            mfr.nPos = i;
                            mfr.nLen = 0;
                            mfr.nType = uMyREType.U_MYRE_TYPE_TAG;
                            mfr.strValue = "</i>";

                            colFormat.Add(mfr);

                            bitalic = ctformatStates.nctNone;
                        }
                        //---------------------

                        //---------------------
                        // strikeout
                        if ((cf.dwEffects & CFM_STRIKEOUT) == CFM_STRIKEOUT)
                        {
                            if (bstrikeout == ctformatStates.nctNone)
                                bstrikeout = ctformatStates.nctNew;
                            else
                                bstrikeout = ctformatStates.nctContinue;
                        }
                        else
                        {
                            if (bstrikeout != ctformatStates.nctNone)
                                bstrikeout = ctformatStates.nctReset;
                        }

                        if (bstrikeout == ctformatStates.nctNew)
                        {
                            mfr = new cMyREFormat();

                            mfr.nPos = i;
                            mfr.nLen = 0;
                            mfr.nType = uMyREType.U_MYRE_TYPE_TAG;
                            mfr.strValue = "<s>";

                            colFormat.Add(mfr);
                        }
                        else if (bstrikeout == ctformatStates.nctReset)
                        {
                            mfr = new cMyREFormat();

                            mfr.nPos = i;
                            mfr.nLen = 0;
                            mfr.nType = uMyREType.U_MYRE_TYPE_TAG;
                            mfr.strValue = "</s>";

                            colFormat.Add(mfr);

                            bstrikeout = ctformatStates.nctNone;
                        }
                        //---------------------

                        //---------------------
                        // underline ?
                        if ((cf.dwEffects & CFE_UNDERLINE) == CFE_UNDERLINE)
                        {
                            if (bunderline == ctformatStates.nctNone)
                                bunderline = ctformatStates.nctNew;
                            else
                                bunderline = ctformatStates.nctContinue;
                        }
                        else
                        {
                            if (bunderline != ctformatStates.nctNone)
                                bunderline = ctformatStates.nctReset;
                        }

                        if (bunderline == ctformatStates.nctNew)
                        {
                            mfr = new cMyREFormat();

                            mfr.nPos = i;
                            mfr.nLen = 0;
                            mfr.nType = uMyREType.U_MYRE_TYPE_TAG;
                            mfr.strValue = "<u>";

                            colFormat.Add(mfr);
                        }
                        else if (bunderline == ctformatStates.nctReset)
                        {
                            mfr = new cMyREFormat();

                            mfr.nPos = i;
                            mfr.nLen = 0;
                            mfr.nType = uMyREType.U_MYRE_TYPE_TAG;
                            mfr.strValue = "</u>";

                            colFormat.Add(mfr);

                            bunderline = ctformatStates.nctNone;
                        }
                        //---------------------

                        //---------------------
                        // superscript ?
                        if ((cf.dwEffects & CFE_SUPERSCRIPT) == CFE_SUPERSCRIPT)
                        {
                            if (super == ctformatStates.nctNone)
                                super = ctformatStates.nctNew;
                            else
                                super = ctformatStates.nctContinue;
                        }
                        else
                        {
                            if (super != ctformatStates.nctNone)
                                super = ctformatStates.nctReset;
                        }

                        if (super == ctformatStates.nctNew)
                        {
                            mfr = new cMyREFormat();

                            mfr.nPos = i;
                            mfr.nLen = 0;
                            mfr.nType = uMyREType.U_MYRE_TYPE_TAG;
                            mfr.strValue = "<sup>";

                            colFormat.Add(mfr);
                        }
                        else if (super == ctformatStates.nctReset)
                        {
                            mfr = new cMyREFormat();

                            mfr.nPos = i;
                            mfr.nLen = 0;
                            mfr.nType = uMyREType.U_MYRE_TYPE_TAG;
                            mfr.strValue = "</sup>";

                            colFormat.Add(mfr);

                            super = ctformatStates.nctNone;
                        }
                        //---------------------

                        //---------------------
                        // subscript ?
                        if ((cf.dwEffects & CFE_SUBSCRIPT) == CFE_SUBSCRIPT)
                        {
                            if (sub == ctformatStates.nctNone)
                                sub = ctformatStates.nctNew;
                            else
                                sub = ctformatStates.nctContinue;
                        }
                        else
                        {
                            if (sub != ctformatStates.nctNone)
                                sub = ctformatStates.nctReset;
                        }

                        if (sub == ctformatStates.nctNew)
                        {
                            mfr = new cMyREFormat();

                            mfr.nPos = i;
                            mfr.nLen = 0;
                            mfr.nType = uMyREType.U_MYRE_TYPE_TAG;
                            mfr.strValue = "<sub>";

                            colFormat.Add(mfr);
                        }
                        else if (sub == ctformatStates.nctReset)
                        {
                            mfr = new cMyREFormat();

                            mfr.nPos = i;
                            mfr.nLen = 0;
                            mfr.nType = uMyREType.U_MYRE_TYPE_TAG;
                            mfr.strValue = "</sub>";

                            colFormat.Add(mfr);

                            sub = ctformatStates.nctNone;
                        }
                        //---------------------
                    }

                    strT += strChar;
                }

                if (bHTML)
                {
                    // close pending tags
                    if (bold != ctformatStates.nctNone)
                    {
                        mfr = new cMyREFormat();

                        mfr.nPos = i;
                        mfr.nLen = 0;
                        mfr.nType = uMyREType.U_MYRE_TYPE_TAG;
                        mfr.strValue = "</b>";

                        colFormat.Add(mfr);
                        //strT += "</b>";
                    }

                    if (bitalic != ctformatStates.nctNone)
                    {
                        mfr = new cMyREFormat();

                        mfr.nPos = i;
                        mfr.nLen = 0;
                        mfr.nType = uMyREType.U_MYRE_TYPE_TAG;
                        mfr.strValue = "</i>";

                        colFormat.Add(mfr);
                        //strT += "</i>";
                    }

                    if (bstrikeout != ctformatStates.nctNone)
                    {
                        mfr = new cMyREFormat();

                        mfr.nPos = i;
                        mfr.nLen = 0;
                        mfr.nType = uMyREType.U_MYRE_TYPE_TAG;
                        mfr.strValue = "</s>";

                        colFormat.Add(mfr);

                        //strT += "</s>";
                    }

                    if (bunderline != ctformatStates.nctNone)
                    {
                        mfr = new cMyREFormat();

                        mfr.nPos = i;
                        mfr.nLen = 0;
                        mfr.nType = uMyREType.U_MYRE_TYPE_TAG;
                        mfr.strValue = "</u>";

                        colFormat.Add(mfr);

                        //strT += "</u>";
                    }

                    if (super != ctformatStates.nctNone)
                    {
                        mfr = new cMyREFormat();

                        mfr.nPos = i;
                        mfr.nLen = 0;
                        mfr.nType = uMyREType.U_MYRE_TYPE_TAG;
                        mfr.strValue = "</sup>";

                        colFormat.Add(mfr);
                        //strT += "</sup>";
                    }

                    if (sub != ctformatStates.nctNone)
                    {
                        mfr = new cMyREFormat();

                        mfr.nPos = i;
                        mfr.nLen = 0;
                        mfr.nType = uMyREType.U_MYRE_TYPE_TAG;
                        mfr.strValue = "</sub>";

                        colFormat.Add(mfr);
                        //strT += "</sub>";
                    }

                    if (strFont != "")
                    {
                        // close pending font format
                        mfr = new cMyREFormat();

                        mfr.nPos = i;
                        mfr.nLen = 0;
                        mfr.nType = uMyREType.U_MYRE_TYPE_TAG;
                        mfr.strValue = "</font>";

                        colFormat.Add(mfr);
                    }
                }

                //--------------------------
                // now, reorder the formatting array
                k = colFormat.Count;
                for (i = 0; i < k - 1; i++)
                {
                    for (int j = i + 1; j < k; j++)
                    {
                        mfr = (cMyREFormat)colFormat[i];
                        cMyREFormat mfr2 = (cMyREFormat)colFormat[j];

                        if (mfr2.nPos < mfr.nPos)
                        {
                            colFormat.RemoveAt(j);
                            colFormat.Insert(i, mfr2);
                            j--;
                        }
                        else if ((mfr2.nPos == mfr.nPos) && (mfr2.nLen < mfr.nLen))
                        {
                            colFormat.RemoveAt(j);
                            colFormat.Insert(i, mfr2);
                            j--;
                        }
                    }
                }
                //--------------------------


                //--------------------------
                // apply format by replacing and inserting HTML tags
                // stored in the Format Array
                int nAcum = 0;
                for (i = 0; i < k; i++)
                {
                    mfr = (cMyREFormat)colFormat[i];

                    strHTML += strT.Substring(nAcum, mfr.nPos - nAcum) + mfr.strValue;
                    nAcum = mfr.nPos + mfr.nLen;
                }

                if (nAcum < strT.Length)
                    strHTML += strT.Substring(nAcum);
                //--------------------------
            }
            catch
            {
                //MessageBox.Show(ex.Message);
            }
            finally
            {
                //--------------------------
                // finish, restore
                this.SelectionStart = nStart;
                this.SelectionLength = nEnd;

                this.EndUpdate();
                this.HideSelection = false;
                //--------------------------
            }

            return strHTML;
        }
        #endregion

        #region "base64 images"
        public ArrayList images = new ArrayList();

        public void InsertImage(string FileName)
        {
            string rtf_data = "";
            try
            {
                // If file is an icon
                if (FileName.ToUpper().EndsWith(".ICO"))
                {
                    // Create a new icon, get it's handle, and create a bitmap from
                    // its handle
                    rtf_data = this.InsertImage(Bitmap.FromHicon((new Icon(FileName)).Handle));
                }
                else
                {
                    // Create a bitmap from the filename
                    rtf_data = this.InsertImage(Image.FromFile(FileName));
                }
            }
            catch (Exception _e)
            {
                MessageBox.Show("The file could not be opened:\n\n" + _e.Message, "File Open Error");
            }

            BitmapBase64 myBitmap = new BitmapBase64(FileName, rtf_data);
            images.Add(myBitmap);
        }
        public void InsertImage(string base64, string type)
        {
            BitmapBase64 myBitmap = new BitmapBase64(base64, type, true);
            string rtf_data = "";
            try
            {
                    rtf_data = this.InsertImage(myBitmap.bitmapImage);
            }
            catch (Exception _e)
            {
                MessageBox.Show("The file could not be opened:\n\n" + _e.Message, "File Open Error");
            }
            myBitmap.RTFstring = rtf_data.ToLower(); ;
            images.Add(myBitmap);
        }


        
        #endregion

        #region "Add HTML text"
        public void AddHTML(string strHTML)
        {
            CHARFORMAT cf;
            PARAFORMAT pf;

            cf = this.DefaultCharFormat; // to apply character formatting
            pf = this.DefaultParaFormat; // to apply paragraph formatting

            char[] chtrim = { ' ', '\x0000' };

            this.HideSelection = true;
            this.BeginUpdate();

            try
            {
                // process text
                while (strHTML.Length > 0)
                {
                    string strData = strHTML;

                reinit:

                    // looking for start tags
                    int nStart = strHTML.IndexOf('<');
                    if (nStart >= 0)
                    {
                        if (nStart > 0)
                        {
                            // tag is not the first character, so
                            // we need to add text to control and continue
                            // looking for tags at the begining of the text
                            strData = strHTML.Substring(0, nStart);
                            strHTML = strHTML.Substring(nStart);
                        }
                        else
                        {
                            // ok, get tag value
                            int nEnd = strHTML.IndexOf('>', nStart);
                            if (nEnd > nStart)
                            {
                                if ((nEnd - nStart) > 0)
                                {
                                    string strTag = strHTML.Substring(nStart, nEnd - nStart + 1);  
                                    strTag = strTag.ToLower();

                                    if (strTag == "<b>")
                                    {
                                        cf.dwMask |= CFM_WEIGHT | CFM_BOLD;
                                        cf.dwEffects |= CFE_BOLD;
                                        cf.wWeight = FW_BOLD;
                                    }
                                    else if (strTag == "<i>")
                                    {
                                        cf.dwMask |= CFM_ITALIC;
                                        cf.dwEffects |= CFE_ITALIC;
                                    }
                                    else if (strTag == "<u>")
                                    {
                                        cf.dwMask |= CFM_UNDERLINE | CFM_UNDERLINETYPE;
                                        cf.dwEffects |= CFE_UNDERLINE;
                                        cf.bUnderlineType = CFU_UNDERLINE;
                                    }
                                    else if (strTag == "<s>")
                                    {
                                        cf.dwMask |= CFM_STRIKEOUT;
                                        cf.dwEffects |= CFE_STRIKEOUT;
                                    }
                                    else if (strTag == "<sup>")
                                    {
                                        cf.dwMask |= CFM_SUPERSCRIPT;
                                        cf.dwEffects |= CFE_SUPERSCRIPT;
                                    }
                                    else if (strTag == "<sub>")
                                    {
                                        cf.dwMask |= CFM_SUBSCRIPT;
                                        cf.dwEffects |= CFE_SUBSCRIPT;
                                    }
                                    else if ((strTag.Length > 4) && (strTag.Substring(0, 4) == "<img"))
                                    {
                                        string strTag_noCaseChange = strHTML.Substring(nStart, nEnd - nStart + 1); 
                                        string type = "";
                                        string base64 = "";
                                        int index_type = strTag.IndexOf("data:image");
                                        if (index_type > 0)
                                        {
                                            int index_endtype = strTag.IndexOf(";", index_type);
                                            if (index_endtype -11 > index_type)
                                            {
                                                type = strTag.Substring(index_type + 11, index_endtype - index_type - 11);
                                            }
                                        }
                                        int index_base64 = strTag.IndexOf("base64,");
                                        if (index_base64 - 7 > 0)
                                        {
                                            int index_endbase64 = strTag.IndexOf("\"", index_base64);
                                            if (index_endbase64 > index_base64)
                                            {
                                                base64 = strTag_noCaseChange.Substring(index_base64 + 7, index_endbase64 - index_base64 - 7);
                                            }
                                        }
                                        this.InsertImage(base64, type);
                                    }
                                    else if ((strTag.Length > 2) && (strTag.Substring(0, 2) == "<p"))
                                    {
                                        if (strTag.IndexOf("align=\"left\"") > 0)
                                        {
                                            pf.dwMask |= PFM_ALIGNMENT;
                                            pf.wAlignment = (short)PFA_LEFT;
                                        }
                                        else if (strTag.IndexOf("align=\"right\"") > 0)
                                        {
                                            pf.dwMask |= PFM_ALIGNMENT;
                                            pf.wAlignment = (short)PFA_RIGHT;
                                        }
                                        else if (strTag.IndexOf("align=\"center\"") > 0)
                                        {
                                            pf.dwMask |= PFM_ALIGNMENT;
                                            pf.wAlignment = (short)PFA_CENTER;
                                        }
                                    }
                                    else if ((strTag.Length > 5) && (strTag.Substring(0, 5) == "<font"))
                                    {
                                        string strFont = new string(cf.szFaceName);
                                        strFont = strFont.Trim(chtrim);
                                        int crFont = cf.crTextColor;
                                        int yHeight = cf.yHeight;

                                        int nFace = strTag.IndexOf("face=");
                                        if (nFace > 0)
                                        {
                                            int nFaceEnd = strTag.IndexOf('\"', nFace + 6);
                                            if (nFaceEnd > nFace)
                                                strFont = strTag.Substring(nFace + 6, nFaceEnd - nFace - 6);
                                        }

                                        int nSize = strTag.IndexOf("size=");
                                        if (nSize > 0)
                                        {
                                            int nSizeEnd = strTag.IndexOf('\"', nSize + 6);
                                            if (nSizeEnd > nSize)
                                            {
                                                yHeight = int.Parse(strTag.Substring(nSize + 6, nSizeEnd - nSize - 6));
                                                yHeight *= (20 * 5);
                                            }
                                        }

                                        int nColor = strTag.IndexOf("color=");
                                        if (nColor > 0)
                                        {
                                            int nColorEnd = strTag.IndexOf('\"', nColor + 7);
                                            if (nColorEnd > nColor)
                                            {
                                                if (strTag.Substring(nColor + 7, 1) == "#")
                                                {
                                                    string strCr = strTag.Substring(nColor + 8, nColorEnd - nColor - 8);
                                                    int nCr = Convert.ToInt32(strCr, 16);

                                                    Color color = Color.FromArgb(nCr);

                                                    crFont = GetCOLORREF(color);
                                                }
                                                else
                                                {
                                                    crFont = int.Parse(strTag.Substring(nColor + 7, nColorEnd - nColor - 7));
                                                }
                                            }
                                        }

                                        cf.szFaceName = new char[LF_FACESIZE];
                                        strFont.CopyTo(0, cf.szFaceName, 0, Math.Min(LF_FACESIZE - 1, strFont.Length));
                                        //cf.szFaceName = strFont.ToCharArray(0, Math.Min(strFont.Length, LF_FACESIZE));
                                        cf.crTextColor = crFont;
                                        cf.yHeight = yHeight;

                                        cf.dwMask |= CFM_COLOR | CFM_SIZE | CFM_FACE;
                                        cf.dwEffects &= ~CFE_AUTOCOLOR;
                                    }
                                    else if (strTag == "<li>")
                                    {
                                        if (pf.wNumbering != PFN_BULLET)
                                        {
                                            pf.dwMask |= PFM_NUMBERING;
                                            pf.wNumbering = (short)PFN_BULLET;
                                        }
                                    }
                                    else if (strTag == "</b>")
                                    {
                                        cf.dwEffects &= ~CFE_BOLD;
                                        cf.wWeight = FW_NORMAL;
                                    }
                                    else if (strTag == "</i>")
                                    {
                                        cf.dwEffects &= ~CFE_ITALIC;
                                    }
                                    else if (strTag == "</u>")
                                    {
                                        cf.dwEffects &= ~CFE_UNDERLINE;
                                    }
                                    else if (strTag == "</s>")
                                    {
                                        cf.dwEffects &= ~CFM_STRIKEOUT;
                                    }
                                    else if (strTag == "</sup>")
                                    {
                                        cf.dwEffects &= ~CFE_SUPERSCRIPT;
                                    }
                                    else if (strTag == "</sub>")
                                    {
                                        cf.dwEffects &= ~CFE_SUBSCRIPT;
                                    }
                                    else if (strTag == "</font>")
                                    {
                                        //Set Defaults
                                        cf.szFaceName = new char[LF_FACESIZE];
                                        this.Font.FontFamily.Name.CopyTo(0, cf.szFaceName, 0, Math.Min(LF_FACESIZE - 1, this.Font.FontFamily.Name.Length));
                                        cf.crTextColor = GetCOLORREF(0, 0, 0);
                                        cf.yHeight = this.DefaultCharFormat.yHeight;
                                    }
                                    else if (strTag == "</p>")
                                    {
                                    }
                                    else if (strTag == "</li>")
                                    {
                                    }

                                    //-------------------------------
                                    // now, remove tag from HTML
                                    int nStart2 = strHTML.IndexOf("<", nEnd + 1);
                                    if (nStart2 > 0)
                                    {
                                        // extract partial data
                                        strData = strHTML.Substring(nEnd + 1, nStart2 - nEnd - 1);
                                        strHTML = strHTML.Substring(nStart2);
                                    }
                                    else
                                    {
                                        // get remain text and finish
                                        if ((nEnd + 1) < strHTML.Length)
                                            strData = strHTML.Substring(nEnd + 1);
                                        else
                                            strData = "";

                                        strHTML = "";
                                    }
                                    //-------------------------------s


                                    //-------------------------------
                                    // have we any continuos tag ?
                                    if (strData.Length > 0)
                                    {
                                        // yes, ok, goto to reinit
                                        if (strData[0] == '<')
                                            goto reinit;
                                    }
                                    //-------------------------------
                                }
                                else
                                {
                                    // we have not found any valid tag
                                    strHTML = "";
                                }
                            }
                            else
                            {
                                // we have not found any valid tag
                                strHTML = "";
                            }
                        }
                    }
                    else
                    {
                        // we have not found any tag
                        strHTML = "";
                    }

                    if (strData.Length > 0)
                    {
                        //-------------------------------
                        // replace entities
                        strData = strData.Replace("&amp;", "&");
                        strData = strData.Replace("&lt;", "<");
                        strData = strData.Replace("&gt;", ">");
                        strData = strData.Replace("&apos;", "'");
                        strData = strData.Replace("&quot;", "\"");
                        //-------------------------------

                        string strAux = strData; // use another copy

                        while (strAux.Length > 0)
                        {
                            //-----------------------
                            int nLen = strAux.Length;
                            //-----------------------

                            //-------------------------------
                            // now, add text to control
                            int nStartCache = this.SelectionStart;
                            string strt = strAux.Substring(0, nLen);

                            this.SelectedText = strt;
                            strAux = strAux.Remove(0, nLen);

                            this.SelectionStart = nStartCache;
                            this.SelectionLength = strt.Length;
                            //-------------------------------

                            //-------------------------------
                            // apply format
                            this.ParaFormat = pf;
                            this.CharFormat = cf;
                            //-------------------------------

                            //fix font family problem
                            string strEndFont = new string(cf.szFaceName);
                            if (strEndFont != this.Font.FontFamily.Name)
                            {
                                Font nFont = new Font(strEndFont, this.SelectionFont.Size, this.SelectionFont.Style);
                                this.SelectionFont = nFont;
                            }
                            // reposition to final
                            this.SelectionStart = this.TextLength + 1;
                            this.SelectionLength = 0;
                        }

                        // reposition to final
                        this.SelectionStart = this.TextLength + 1;
                        this.SelectionLength = 0;

                        //-------------------------------
                        // new paragraph requires to reset alignment
                        if ((strData.IndexOf("\r\n", 0) >= 0) || (strData.IndexOf("\n", 0) >= 0))
                        {
                            pf.dwMask = PFM_ALIGNMENT | PFM_NUMBERING;
                            pf.wAlignment = (short)PFA_LEFT;
                            pf.wNumbering = 0;
                        }
                        //-------------------------------
                    }
                } // while (strHTML.Length > 0)
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message);
            }
            finally
            {
                // reposition to final
                this.SelectionStart = this.TextLength + 1;
                this.SelectionLength = 0;

                this.EndUpdate();
                this.HideSelection = false;
            }
        }
        #endregion

		#region My Enums

		// Specifies the flags/options for the unmanaged call to the GDI+ method
		// Metafile.EmfToWmfBits().
		private enum EmfToWmfBitsFlags {

			// Use the default conversion
			EmfToWmfBitsFlagsDefault = 0x00000000,

			// Embedded the source of the EMF metafiel within the resulting WMF
			// metafile
			EmfToWmfBitsFlagsEmbedEmf = 0x00000001,

			// Place a 22-byte header in the resulting WMF file.  The header is
			// required for the metafile to be considered placeable.
			EmfToWmfBitsFlagsIncludePlaceable = 0x00000002,

			// Don't simulate clipping by using the XOR operator.
			EmfToWmfBitsFlagsNoXORClip = 0x00000004
		};

		#endregion

		#region My Structs

		// Definitions for colors in an RTF document
		private struct RtfColorDef {
			public const string Black = @"\red0\green0\blue0";
			public const string Maroon = @"\red128\green0\blue0";
			public const string Green = @"\red0\green128\blue0";
			public const string Olive = @"\red128\green128\blue0";
			public const string Navy = @"\red0\green0\blue128";
			public const string Purple = @"\red128\green0\blue128";
			public const string Teal = @"\red0\green128\blue128";
			public const string Gray = @"\red128\green128\blue128";
			public const string Silver = @"\red192\green192\blue192";
			public const string Red = @"\red255\green0\blue0";
			public const string Lime = @"\red0\green255\blue0";
			public const string Yellow = @"\red255\green255\blue0";
			public const string Blue = @"\red0\green0\blue255";
			public const string Fuchsia = @"\red255\green0\blue255";
			public const string Aqua = @"\red0\green255\blue255";
			public const string White = @"\red255\green255\blue255";
		}

		// Control words for RTF font families
		private struct RtfFontFamilyDef {
			public const string Unknown = @"\fnil";
			public const string Roman = @"\froman";
			public const string Swiss = @"\fswiss";
			public const string Modern = @"\fmodern";
			public const string Script = @"\fscript";
			public const string Decor = @"\fdecor";
			public const string Technical = @"\ftech";
			public const string BiDirect = @"\fbidi";
		}

		#endregion

		#region My Constants

		// Not used in this application.  Descriptions can be found with documentation
		// of Windows GDI function SetMapMode
		private const int MM_TEXT = 1;
		private const int MM_LOMETRIC = 2;
		private const int MM_HIMETRIC = 3;
		private const int MM_LOENGLISH = 4;
		private const int MM_HIENGLISH = 5;
		private const int MM_TWIPS = 6;

		// Ensures that the metafile maintains a 1:1 aspect ratio
		private const int MM_ISOTROPIC = 7;

		// Allows the x-coordinates and y-coordinates of the metafile to be adjusted
		// independently
		private const int MM_ANISOTROPIC = 8;

		// Represents an unknown font family
		private const string FF_UNKNOWN = "UNKNOWN";

		// The number of hundredths of millimeters (0.01 mm) in an inch
		// For more information, see GetImagePrefix() method.
		private const int HMM_PER_INCH = 2540;

		// The number of twips in an inch
		// For more information, see GetImagePrefix() method.
		private const int TWIPS_PER_INCH = 1440;

		#endregion

		#region My Privates

		// The default text color
		private RtfColor textColor;

		// The default text background color
		private RtfColor highlightColor;

		// Dictionary that maps color enums to RTF color codes
		private HybridDictionary rtfColor;

		// Dictionary that mapas Framework font families to RTF font families
		private HybridDictionary rtfFontFamily;

		// The horizontal resolution at which the control is being displayed
		private float xDpi;

		// The vertical resolution at which the control is being displayed
		private float yDpi;

		#endregion

		#region Elements required to create an RTF document
		
		/* RTF HEADER
		 * ----------
		 * 
		 * \rtf[N]		- For text to be considered to be RTF, it must be enclosed in this tag.
		 *				  rtf1 is used because the RichTextBox conforms to RTF Specification
		 *				  version 1.
		 * \ansi		- The character set.
		 * \ansicpg[N]	- Specifies that unicode characters might be embedded. ansicpg1252
		 *				  is the default used by Windows.
		 * \deff[N]		- The default font. \deff0 means the default font is the first font
		 *				  found.
		 * \deflang[N]	- The default language. \deflang1033 specifies US English.
		 * */
		private const string RTF_HEADER = @"{\rtf1\ansi\ansicpg1252\deff0\deflang1033";

		/* RTF DOCUMENT AREA
		 * -----------------
		 * 
		 * \viewkind[N]	- The type of view or zoom level.  \viewkind4 specifies normal view.
		 * \uc[N]		- The number of bytes corresponding to a Unicode character.
		 * \pard		- Resets to default paragraph properties
		 * \cf[N]		- Foreground color.  \cf1 refers to the color at index 1 in
		 *				  the color table
		 * \f[N]		- Font number. \f0 refers to the font at index 0 in the font
		 *				  table.
		 * \fs[N]		- Font size in half-points.
		 * */
		private const string RTF_DOCUMENT_PRE = @"\viewkind4\uc1\pard\cf1\f0\fs20";
		private const string RTF_DOCUMENT_POST = @"\cf0\fs17}";
		private string RTF_IMAGE_POST = @"}";

		#endregion

		#region Accessors

		// TODO: This can be ommitted along with RemoveBadCharacters
		// Overrides the default implementation of RTF.  This is done because the control
		// was originally developed to run in an instant messenger that uses the
		// Jabber XML-based protocol.  The framework would throw an exception when the
		// XML contained the null character, so I filtered out.
		public new string Rtf {
			get {return RemoveBadChars(base.Rtf);}
			set {base.Rtf = value;}
		}

		// The color of the text
		public RtfColor TextColor {
			get {return textColor;}
			set {textColor = value;}
		}

		// The color of the highlight
		public RtfColor HiglightColor {
			get {return highlightColor;}
			set {highlightColor = value;}
		}

		#endregion

		#region Constructors

		/// <summary>
		/// Initializes the text colors, creates dictionaries for RTF colors and
		/// font families, and stores the horizontal and vertical resolution of
		/// the RichTextBox's graphics context.
		/// </summary>
                public HtmlRichTextBox() : base() 
    {

			// Initialize default text and background colors
			textColor = RtfColor.Black;
			highlightColor = RtfColor.White;

			// Initialize the dictionary mapping color codes to definitions
			rtfColor = new HybridDictionary();
			rtfColor.Add(RtfColor.Aqua, RtfColorDef.Aqua);
			rtfColor.Add(RtfColor.Black, RtfColorDef.Black);
			rtfColor.Add(RtfColor.Blue, RtfColorDef.Blue);
			rtfColor.Add(RtfColor.Fuchsia, RtfColorDef.Fuchsia);
			rtfColor.Add(RtfColor.Gray, RtfColorDef.Gray);
			rtfColor.Add(RtfColor.Green, RtfColorDef.Green);
			rtfColor.Add(RtfColor.Lime, RtfColorDef.Lime);
			rtfColor.Add(RtfColor.Maroon, RtfColorDef.Maroon);
			rtfColor.Add(RtfColor.Navy, RtfColorDef.Navy);
			rtfColor.Add(RtfColor.Olive, RtfColorDef.Olive);
			rtfColor.Add(RtfColor.Purple, RtfColorDef.Purple);
			rtfColor.Add(RtfColor.Red, RtfColorDef.Red);
			rtfColor.Add(RtfColor.Silver, RtfColorDef.Silver);
			rtfColor.Add(RtfColor.Teal, RtfColorDef.Teal);
			rtfColor.Add(RtfColor.White, RtfColorDef.White);
			rtfColor.Add(RtfColor.Yellow, RtfColorDef.Yellow);

			// Initialize the dictionary mapping default Framework font families to
			// RTF font families
			rtfFontFamily = new HybridDictionary();
			rtfFontFamily.Add(FontFamily.GenericMonospace.Name, RtfFontFamilyDef.Modern);
			rtfFontFamily.Add(FontFamily.GenericSansSerif, RtfFontFamilyDef.Swiss);
			rtfFontFamily.Add(FontFamily.GenericSerif, RtfFontFamilyDef.Roman);
			rtfFontFamily.Add(FF_UNKNOWN, RtfFontFamilyDef.Unknown);

			// Get the horizontal and vertical resolutions at which the object is
			// being displayed
			using(Graphics _graphics = this.CreateGraphics()) {
				xDpi = _graphics.DpiX;
				yDpi = _graphics.DpiY;
			}
}
		/// <summary>
		/// Calls the default constructor then sets the text color.
		/// </summary>
		/// <param name="_textColor"></param>
		public HtmlRichTextBox(RtfColor _textColor) : this() {
			textColor = _textColor;
		}

		/// <summary>
		/// Calls the default constructor then sets te text and highlight colors.
		/// </summary>
		/// <param name="_textColor"></param>
		/// <param name="_highlightColor"></param>
        public HtmlRichTextBox(RtfColor _textColor, RtfColor _highlightColor)
            : this()
        {
			textColor = _textColor;
			highlightColor = _highlightColor;
		}

		#endregion

		#region Append RTF or Text to RichTextBox Contents

		/// <summary>
		/// Assumes the string passed as a paramter is valid RTF text and attempts
		/// to append it as RTF to the content of the control.
		/// </summary>
		/// <param name="_rtf"></param>
		public void AppendRtf(string _rtf) {

			// Move caret to the end of the text
			this.Select(this.TextLength, 0);

			// Since SelectedRtf is null, this will append the string to the
			// end of the existing RTF
			this.SelectedRtf = _rtf;
		}

		/// <summary>
		/// Assumes that the string passed as a parameter is valid RTF text and
		/// attempts to insert it as RTF into the content of the control.
		/// </summary>
		/// <remarks>
		/// NOTE: The text is inserted wherever the caret is at the time of the call,
		/// and if any text is selected, that text is replaced.
		/// </remarks>
		/// <param name="_rtf"></param>
		public void InsertRtf(string _rtf) {
			this.SelectedRtf = _rtf;
		}

		/// <summary>
		/// Appends the text using the current font, text, and highlight colors.
		/// </summary>
		/// <param name="_text"></param>
		public void AppendTextAsRtf(string _text) {
			AppendTextAsRtf(_text, this.Font);
		}


		/// <summary>
		/// Appends the text using the given font, and current text and highlight
		/// colors.
		/// </summary>
		/// <param name="_text"></param>
		/// <param name="_font"></param>
		public void AppendTextAsRtf(string _text, Font _font) {
			AppendTextAsRtf(_text, _font, textColor);
		}
		
		/// <summary>
		/// Appends the text using the given font and text color, and the current
		/// highlight color.
		/// </summary>
		/// <param name="_text"></param>
		/// <param name="_font"></param>
		/// <param name="_color"></param>
		public void AppendTextAsRtf(string _text, Font _font, RtfColor _textColor) {
			AppendTextAsRtf(_text, _font, _textColor, highlightColor);
		}

		/// <summary>
		/// Appends the text using the given font, text, and highlight colors.  Simply
		/// moves the caret to the end of the RichTextBox's text and makes a call to
		/// insert.
		/// </summary>
		/// <param name="_text"></param>
		/// <param name="_font"></param>
		/// <param name="_textColor"></param>
		/// <param name="_backColor"></param>
		public void AppendTextAsRtf(string _text, Font _font, RtfColor _textColor, RtfColor _backColor) {
			// Move carret to the end of the text
			this.Select(this.TextLength, 0);

			InsertTextAsRtf(_text, _font, _textColor, _backColor);
		}

		#endregion

		#region Insert Plain Text

		/// <summary>
		/// Inserts the text using the current font, text, and highlight colors.
		/// </summary>
		/// <param name="_text"></param>
		public void InsertTextAsRtf(string _text) {
			InsertTextAsRtf(_text, this.Font);
		}


		/// <summary>
		/// Inserts the text using the given font, and current text and highlight
		/// colors.
		/// </summary>
		/// <param name="_text"></param>
		/// <param name="_font"></param>
		public void InsertTextAsRtf(string _text, Font _font) {
			InsertTextAsRtf(_text, _font, textColor);
		}
		
		/// <summary>
		/// Inserts the text using the given font and text color, and the current
		/// highlight color.
		/// </summary>
		/// <param name="_text"></param>
		/// <param name="_font"></param>
		/// <param name="_color"></param>
		public void InsertTextAsRtf(string _text, Font _font, RtfColor _textColor) {
			InsertTextAsRtf(_text, _font, _textColor, highlightColor);
		}

		/// <summary>
		/// Inserts the text using the given font, text, and highlight colors.  The
		/// text is wrapped in RTF codes so that the specified formatting is kept.
		/// You can only assign valid RTF to the RichTextBox.Rtf property, else
		/// an exception is thrown.  The RTF string should follow this format ...
		/// 
		/// {\rtf1\ansi\ansicpg1252\deff0\deflang1033{\fonttbl{[FONTS]}{\colortbl ;[COLORS]}}
		/// \viewkind4\uc1\pard\cf1\f0\fs20 [DOCUMENT AREA] }
		/// 
		/// </summary>
		/// <remarks>
		/// NOTE: The text is inserted wherever the caret is at the time of the call,
		/// and if any text is selected, that text is replaced.
		/// </remarks>
		/// <param name="_text"></param>
		/// <param name="_font"></param>
		/// <param name="_color"></param>
		/// <param name="_color"></param>
		public void InsertTextAsRtf(string _text, Font _font, RtfColor _textColor, RtfColor _backColor) {

			StringBuilder _rtf = new StringBuilder();

			// Append the RTF header
			_rtf.Append(RTF_HEADER);

			// Create the font table from the font passed in and append it to the
			// RTF string
			_rtf.Append(GetFontTable(_font));

			// Create the color table from the colors passed in and append it to the
			// RTF string
			_rtf.Append(GetColorTable(_textColor, _backColor));

			// Create the document area from the text to be added as RTF and append
			// it to the RTF string.
			_rtf.Append(GetDocumentArea(_text, _font));

			this.SelectedRtf = _rtf.ToString();
		}

		/// <summary>
		/// Creates the Document Area of the RTF being inserted. The document area
		/// (in this case) consists of the text being added as RTF and all the
		/// formatting specified in the Font object passed in. This should have the
		/// form ...
		/// 
		/// \viewkind4\uc1\pard\cf1\f0\fs20 [DOCUMENT AREA] }
		///
		/// </summary>
		/// <param name="_text"></param>
		/// <param name="_font"></param>
		/// <returns>
		/// The document area as a string.
		/// </returns>
		private string GetDocumentArea(string _text, Font _font) {

			StringBuilder _doc = new StringBuilder();
			
			// Append the standard RTF document area control string
			_doc.Append(RTF_DOCUMENT_PRE);

			// Set the highlight color (the color behind the text) to the
			// third color in the color table.  See GetColorTable for more details.
			_doc.Append(@"\highlight2");

			// If the font is bold, attach corresponding tag
			if (_font.Bold)
				_doc.Append(@"\b");

			// If the font is italic, attach corresponding tag
			if (_font.Italic)
				_doc.Append(@"\i");

			// If the font is strikeout, attach corresponding tag
			if (_font.Strikeout)
				_doc.Append(@"\strike");

			// If the font is underlined, attach corresponding tag
			if (_font.Underline)
				_doc.Append(@"\ul");

			// Set the font to the first font in the font table.
			// See GetFontTable for more details.
			_doc.Append(@"\f0");

			// Set the size of the font.  In RTF, font size is measured in
			// half-points, so the font size is twice the value obtained from
			// Font.SizeInPoints
			_doc.Append(@"\fs");
			_doc.Append((int)Math.Round((2 * _font.SizeInPoints)));

			// Apppend a space before starting actual text (for clarity)
			_doc.Append(@" ");

			// Append actual text, however, replace newlines with RTF \par.
			// Any other special text should be handled here (e.g.) tabs, etc.
			_doc.Append(_text.Replace("\n", @"\par "));

			// RTF isn't strict when it comes to closing control words, but what the
			// heck ...

			// Remove the highlight
			_doc.Append(@"\highlight0");

			// If font is bold, close tag
			if (_font.Bold)
				_doc.Append(@"\b0");

			// If font is italic, close tag
			if (_font.Italic)
				_doc.Append(@"\i0");

			// If font is strikeout, close tag
			if (_font.Strikeout)
				_doc.Append(@"\strike0");

			// If font is underlined, cloes tag
			if (_font.Underline)
				_doc.Append(@"\ulnone");

			// Revert back to default font and size
			_doc.Append(@"\f0");
			_doc.Append(@"\fs20");

			// Close the document area control string
			_doc.Append(RTF_DOCUMENT_POST);

			return _doc.ToString();
		}

		#endregion

		#region Insert Image

		/// <summary>
		/// Inserts an image into the RichTextBox.  The image is wrapped in a Windows
		/// Format Metafile, because although Microsoft discourages the use of a WMF,
		/// the RichTextBox (and even MS Word), wraps an image in a WMF before inserting
		/// the image into a document.  The WMF is attached in HEX format (a string of
		/// HEX numbers).
		/// 
		/// The RTF Specification v1.6 says that you should be able to insert bitmaps,
		/// .jpegs, .gifs, .pngs, and Enhanced Metafiles (.emf) directly into an RTF
		/// document without the WMF wrapper. This works fine with MS Word,
		/// however, when you don't wrap images in a WMF, WordPad and
		/// RichTextBoxes simply ignore them.  Both use the riched20.dll or msfted.dll.
		/// </summary>
		/// <remarks>
		/// NOTE: The image is inserted wherever the caret is at the time of the call,
		/// and if any text is selected, that text is replaced.
		/// </remarks>
		/// <param name="_image"></param>
		public string InsertImage(Image _image) {

			StringBuilder _rtf = new StringBuilder();

			// Append the RTF header
			_rtf.Append(RTF_HEADER);

			// Create the font table using the RichTextBox's current font and append
			// it to the RTF string
			_rtf.Append(GetFontTable(this.Font));

			// Create the image control string and append it to the RTF string
			_rtf.Append(GetImagePrefix(_image));
         
			// Create the Windows Metafile and append its bytes in HEX format
            string rtf_data = GetRtfImage(_image);
            _rtf.Append(rtf_data);

			// Close the RTF image control string
			_rtf.Append(RTF_IMAGE_POST);

			this.SelectedRtf = _rtf.ToString();
            return rtf_data;
		}

		/// <summary>
		/// Creates the RTF control string that describes the image being inserted.
		/// This description (in this case) specifies that the image is an
		/// MM_ANISOTROPIC metafile, meaning that both X and Y axes can be scaled
		/// independently.  The control string also gives the images current dimensions,
		/// and its target dimensions, so if you want to control the size of the
		/// image being inserted, this would be the place to do it. The prefix should
		/// have the form ...
		/// 
		/// {\pict\wmetafile8\picw[A]\pich[B]\picwgoal[C]\pichgoal[D]
		/// 
		/// where ...
		/// 
		/// A	= current width of the metafile in hundredths of millimeters (0.01mm)
		///		= Image Width in Inches * Number of (0.01mm) per inch
		///		= (Image Width in Pixels / Graphics Context's Horizontal Resolution) * 2540
		///		= (Image Width in Pixels / Graphics.DpiX) * 2540
		/// 
		/// B	= current height of the metafile in hundredths of millimeters (0.01mm)
		///		= Image Height in Inches * Number of (0.01mm) per inch
		///		= (Image Height in Pixels / Graphics Context's Vertical Resolution) * 2540
		///		= (Image Height in Pixels / Graphics.DpiX) * 2540
		/// 
		/// C	= target width of the metafile in twips
		///		= Image Width in Inches * Number of twips per inch
		///		= (Image Width in Pixels / Graphics Context's Horizontal Resolution) * 1440
		///		= (Image Width in Pixels / Graphics.DpiX) * 1440
		/// 
		/// D	= target height of the metafile in twips
		///		= Image Height in Inches * Number of twips per inch
		///		= (Image Height in Pixels / Graphics Context's Horizontal Resolution) * 1440
		///		= (Image Height in Pixels / Graphics.DpiX) * 1440
		///	
		/// </summary>
		/// <remarks>
		/// The Graphics Context's resolution is simply the current resolution at which
		/// windows is being displayed.  Normally it's 96 dpi, but instead of assuming
		/// I just added the code.
		/// 
		/// According to Ken Howe at pbdr.com, "Twips are screen-independent units
		/// used to ensure that the placement and proportion of screen elements in
		/// your screen application are the same on all display systems."
		/// 
		/// Units Used
		/// ----------
		/// 1 Twip = 1/20 Point
		/// 1 Point = 1/72 Inch
		/// 1 Twip = 1/1440 Inch
		/// 
		/// 1 Inch = 2.54 cm
		/// 1 Inch = 25.4 mm
		/// 1 Inch = 2540 (0.01)mm
		/// </remarks>
		/// <param name="_image"></param>
		/// <returns></returns>
		private string GetImagePrefix(Image _image) {

			StringBuilder _rtf = new StringBuilder();

			// Calculate the current width of the image in (0.01)mm
			int picw = (int)Math.Round((_image.Width / xDpi) * HMM_PER_INCH);

			// Calculate the current height of the image in (0.01)mm
			int pich = (int)Math.Round((_image.Height / yDpi) * HMM_PER_INCH);

			// Calculate the target width of the image in twips
			int picwgoal = (int)Math.Round((_image.Width / xDpi) * TWIPS_PER_INCH);

			// Calculate the target height of the image in twips
			int pichgoal = (int)Math.Round((_image.Height / yDpi) * TWIPS_PER_INCH);

			// Append values to RTF string
			_rtf.Append(@"{\pict\wmetafile8");
			_rtf.Append(@"\picw");
			_rtf.Append(picw);
			_rtf.Append(@"\pich");
			_rtf.Append(pich);
			_rtf.Append(@"\picwgoal");
			_rtf.Append(picwgoal);
			_rtf.Append(@"\pichgoal");
			_rtf.Append(pichgoal);
			_rtf.Append(" ");

			return _rtf.ToString();
		}

		/// <summary>
		/// Use the EmfToWmfBits function in the GDI+ specification to convert a 
		/// Enhanced Metafile to a Windows Metafile
		/// </summary>
		/// <param name="_hEmf">
		/// A handle to the Enhanced Metafile to be converted
		/// </param>
		/// <param name="_bufferSize">
		/// The size of the buffer used to store the Windows Metafile bits returned
		/// </param>
		/// <param name="_buffer">
		/// An array of bytes used to hold the Windows Metafile bits returned
		/// </param>
		/// <param name="_mappingMode">
		/// The mapping mode of the image.  This control uses MM_ANISOTROPIC.
		/// </param>
		/// <param name="_flags">
		/// Flags used to specify the format of the Windows Metafile returned
		/// </param>
		[DllImportAttribute("gdiplus.dll")]
		private static extern uint GdipEmfToWmfBits (IntPtr _hEmf, uint _bufferSize,
			byte[] _buffer, int _mappingMode, EmfToWmfBitsFlags _flags);


		/// <summary>
		/// Wraps the image in an Enhanced Metafile by drawing the image onto the
		/// graphics context, then converts the Enhanced Metafile to a Windows
		/// Metafile, and finally appends the bits of the Windows Metafile in HEX
		/// to a string and returns the string.
		/// </summary>
		/// <param name="_image"></param>
		/// <returns>
		/// A string containing the bits of a Windows Metafile in HEX
		/// </returns>
		private string GetRtfImage(Image _image) {

			StringBuilder _rtf = null;

			// Used to store the enhanced metafile
			MemoryStream _stream = null;

			// Used to create the metafile and draw the image
			Graphics _graphics = null;

			// The enhanced metafile
			Metafile _metaFile = null;

			// Handle to the device context used to create the metafile
			IntPtr _hdc;

			try {
				_rtf = new StringBuilder();
				_stream = new MemoryStream();

				// Get a graphics context from the RichTextBox
				using(_graphics = this.CreateGraphics()) {

					// Get the device context from the graphics context
					_hdc = _graphics.GetHdc();

					// Create a new Enhanced Metafile from the device context
					_metaFile = new Metafile(_stream, _hdc);

					// Release the device context
					_graphics.ReleaseHdc(_hdc);
				}

				// Get a graphics context from the Enhanced Metafile
				using(_graphics = Graphics.FromImage(_metaFile)) {

					// Draw the image on the Enhanced Metafile
					_graphics.DrawImage(_image, new Rectangle(0, 0, _image.Width, _image.Height));

				}

				// Get the handle of the Enhanced Metafile
				IntPtr _hEmf = _metaFile.GetHenhmetafile();

				// A call to EmfToWmfBits with a null buffer return the size of the
				// buffer need to store the WMF bits.  Use this to get the buffer
				// size.
				uint _bufferSize = GdipEmfToWmfBits(_hEmf, 0, null, MM_ANISOTROPIC,
					EmfToWmfBitsFlags.EmfToWmfBitsFlagsDefault);

				// Create an array to hold the bits
				byte[] _buffer = new byte[_bufferSize];

				// A call to EmfToWmfBits with a valid buffer copies the bits into the
				// buffer an returns the number of bits in the WMF.  
				uint _convertedSize = GdipEmfToWmfBits(_hEmf, _bufferSize, _buffer, MM_ANISOTROPIC,
					EmfToWmfBitsFlags.EmfToWmfBitsFlagsDefault);

				// Append the bits to the RTF string
				for(int i = 0; i < _buffer.Length; ++i) {
					_rtf.Append(String.Format("{0:X2}", _buffer[i]));
				}

				return _rtf.ToString();
			}
			finally {
				if(_graphics != null)
					_graphics.Dispose();
				if(_metaFile != null)
					_metaFile.Dispose();
				if(_stream != null)
					_stream.Close();
			}
		}
		
		#endregion

		#region RTF Helpers

		/// <summary>
		/// Creates a font table from a font object.  When an Insert or Append 
		/// operation is performed a font is either specified or the default font
		/// is used.  In any case, on any Insert or Append, only one font is used,
		/// thus the font table will always contain a single font.  The font table
		/// should have the form ...
		/// 
		/// {\fonttbl{\f0\[FAMILY]\fcharset0 [FONT_NAME];}
		/// </summary>
		/// <param name="_font"></param>
		/// <returns></returns>
		private string GetFontTable(Font _font) {

			StringBuilder _fontTable = new StringBuilder();

			// Append table control string
			_fontTable.Append(@"{\fonttbl{\f0");
			_fontTable.Append(@"\");
			
			// If the font's family corresponds to an RTF family, append the
			// RTF family name, else, append the RTF for unknown font family.
			if (rtfFontFamily.Contains(_font.FontFamily.Name))
				_fontTable.Append(rtfFontFamily[_font.FontFamily.Name]);
			else
				_fontTable.Append(rtfFontFamily[FF_UNKNOWN]);

			// \fcharset specifies the character set of a font in the font table.
			// 0 is for ANSI.
			_fontTable.Append(@"\fcharset0 ");

			// Append the name of the font
			_fontTable.Append(_font.Name);

			// Close control string
			_fontTable.Append(@";}}");

			return _fontTable.ToString();
		}

		/// <summary>
		/// Creates a font table from the RtfColor structure.  When an Insert or Append
		/// operation is performed, _textColor and _backColor are either specified
		/// or the default is used.  In any case, on any Insert or Append, only three
		/// colors are used.  The default color of the RichTextBox (signified by a
		/// semicolon (;) without a definition), is always the first color (index 0) in
		/// the color table.  The second color is always the text color, and the third
		/// is always the highlight color (color behind the text).  The color table
		/// should have the form ...
		/// 
		/// {\colortbl ;[TEXT_COLOR];[HIGHLIGHT_COLOR];}
		/// 
		/// </summary>
		/// <param name="_textColor"></param>
		/// <param name="_backColor"></param>
		/// <returns></returns>
		private string GetColorTable(RtfColor _textColor, RtfColor _backColor) {

			StringBuilder _colorTable = new StringBuilder();

			// Append color table control string and default font (;)
			_colorTable.Append(@"{\colortbl ;");

			// Append the text color
			_colorTable.Append(rtfColor[_textColor]);
			_colorTable.Append(@";");

			// Append the highlight color
			_colorTable.Append(rtfColor[_backColor]);
			_colorTable.Append(@";}\n");
					
			return _colorTable.ToString();
		}

		/// <summary>
		/// Called by overrided RichTextBox.Rtf accessor.
		/// Removes the null character from the RTF.  This is residue from developing
		/// the control for a specific instant messaging protocol and can be ommitted.
		/// </summary>
		/// <param name="_originalRtf"></param>
		/// <returns>RTF without null character</returns>
		private string RemoveBadChars(string _originalRtf) {			
			return _originalRtf.Replace("\0", "");
		}

		#endregion
    }


    public class BitmapBase64 : Object
    {
        public Bitmap bitmapImage;
        public string Base64string;
        public string RTFstring;
        public string name;
        public string extension;

        public BitmapBase64()
        {
        }

        public BitmapBase64(string FileName, string RTF)
        {
            LoadFile(FileName, RTF);
        }

        public BitmapBase64(string base64, string type, bool CreateImage)
        {
            this.Base64string = base64;
            this.extension = type;
            if (CreateImage)
            {
                LoadFromBase64(base64);
            }
        }

        public void LoadFromBase64(string base64)
        {
            this.bitmapImage = this.decodeImage(this.FixBase64ForImage(base64));
        }

        public void LoadFile(string FileName, string RTF)
        {
            this.bitmapImage = new Bitmap(FileName);
            this.Base64string = EncodeWithString(FileName);
            this.RTFstring = RTF.ToLower();
            this.name = Path.GetFileNameWithoutExtension(FileName);
            this.extension = Path.GetExtension(FileName).Remove(0,1);
        }

        public void LoadImage(string Base64)
        {
            bitmapImage = decodeImage(FixBase64ForImage(Base64string));
            this.Base64string = Base64;
        }

        
        public string EncodeWithString(string inputFileName)
        {
            System.IO.FileStream inFile;
            byte[] binaryData;

            try
            {
                inFile = new System.IO.FileStream(inputFileName,
                                                  System.IO.FileMode.Open,
                                                  System.IO.FileAccess.Read);
                binaryData = new Byte[inFile.Length];
                long bytesRead = inFile.Read(binaryData, 0,
                                            (int)inFile.Length);
                inFile.Close();
            }
            catch (System.Exception exp)
            {
                // Error creating stream or reading from it.
                System.Console.WriteLine("{0}", exp.Message);
                return "";
            }

            // Convert the binary input into Base64 UUEncoded output.
            string base64String;
            try
            {
                base64String =
                   System.Convert.ToBase64String(binaryData,
                                                 0,
                                                 binaryData.Length);
                return base64String;
            }
            catch (System.ArgumentNullException)
            {
                System.Console.WriteLine("Binary data array is null.");
                return "";
            }
        }

        public string FixBase64ForImage(string Image)
        {
            System.Text.StringBuilder sbText =
               new System.Text.StringBuilder(Image, Image.Length);

            sbText.Replace("\r\n", String.Empty);

            sbText.Replace(" ", String.Empty);

            return sbText.ToString();
        }

        public Bitmap decodeImage(string ImageTest)
        {
            if (ImageTest.Length > 0)
            {
                Byte[] bitmapData = new Byte[ImageTest.Length];

                bitmapData = Convert.FromBase64String(FixBase64ForImage
                  (ImageTest));

                System.IO.MemoryStream streamBitmap = new
                  System.IO.MemoryStream(bitmapData);

                return new Bitmap((Bitmap)Image.FromStream(streamBitmap));
            }
            return null;
        }
    }
}